import traceback
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable

import cursefetch
from tqdm import tqdm

from translatools import TranslatoolsMetadata, Paratranz
from translatools.config import update_deprecated_metadata


class Translatools:
    config: TranslatoolsMetadata
    _conf_path: Path

    def __init__(self, config: TranslatoolsMetadata, conf_path: Path):
        self.config = config
        self._conf_path = conf_path

    def __str__(self):
        return "Translatools(" + str(self.config) + ", conf_path=" + str(self._conf_path) + ")"

    def config_path(self):
        return self._conf_path

    def cwd(self):
        return self._conf_path.parent

    def install(self):
        f = cursefetch.get_project_file(str(self.config.project_id), "latest")
        cursefetch.download_project_file(f, "workspace", uncompress=True)

    def sync_to_paratranz(self, client: Paratranz):
        # load existing
        existing = client.get_file_list(self.config.paratranz_id)

        def upload_or_update_files(paths_: Iterable[Path], desc: str):
            for json_path in (bar := tqdm(paths_, desc=desc)):
                try:
                    bar.set_postfix_str(str(json_path))
                    normalized_path = json_path.relative_to(self.cwd())
                    if normalized_path.as_posix() in existing:  # as_posix() makes '\\' to '/' when on Windows
                        data = existing[normalized_path.as_posix()]
                        file_id = data["id"]
                        client.update_file(self.config.paratranz_id, file_id, json_path)
                    else:
                        client.put_file(self.config.paratranz_id, json_path, normalized_path)
                except Exception as e:
                    print(f"Failed to upload {json_path}")
                    traceback.print_exception(e)

        # Migrate!
        update_deprecated_metadata(self.config)

        # upload or update the files from tracked files
        for tracked_file in self.config.tracked_files:
            paths = tracked_file.get_transformed_json_paths(self.cwd())
            upload_or_update_files(paths, f"Sync-ing {tracked_file.path} as {tracked_file.type}")


@dataclass
class FTBQuestKeyGeneratingConfig:
    quest_title: str = "ftbquests.chapter.{chapter_id}.quests.{quest_id}.title"
    quest_subtitle: str = "ftbquests.chapter.{chapter_id}.quests.{quest_id}.subtitle"
    quest_description: str = "ftbquests.chapter.{chapter_id}.quests.{quest_id}.description.{description_index}"
    generate_description_index_for_empty_lines: bool = False

    @staticmethod
    def get_default() -> "FTBQuestKeyGeneratingConfig":
        return FTBQuestKeyGeneratingConfig()

    def get_title_key(self, chapter_id: str, quest_id: str):
        return (self.quest_title
                .replace("{chapter_id}", chapter_id)
                .replace("{quest_id}", quest_id))

    def get_subtitle_key(self, chapter_id: str, quest_id: str):
        return (self.quest_subtitle
                .replace("{chapter_id}", chapter_id)
                .replace("{quest_id}", quest_id))

    def get_description_key(self, chapter_id: str, quest_id: str, description_index: int):
        return (self.quest_description
                .replace("{chapter_id}", chapter_id)
                .replace("{quest_id}", quest_id)
                .replace("{description_index}", str(description_index)))



import dataclasses
import json
from dataclasses import dataclass, field
from enum import StrEnum
from pathlib import Path

import dacite
import ftb_snbt_lib
from ftb_snbt_lib.tag import List, Compound

from translatools.translatools import FTBQuestKeyGeneratingConfig


class FileType(StrEnum):
    """
    The type of the file.
    This describes how to handle the file, like how to extract key-value pairs.
    """
    JSON_KV = "json_kv"
    LANG_KV = "lang_kv"
    # FTB Quests
    FTBQuests_Chapter = "ftbquests_chapter"


@dataclass
class TrackedFile:
    path: str
    type: FileType

    def get_paths(self, cwd: Path) -> list[Path]:
        return list(cwd.glob(self.path))

    def get_transformed_json_paths(self, cwd: Path) -> list[Path]:
        match self.type:
            case FileType.JSON_KV:
                # match files by glob and directly return
                return list(cwd.glob(self.path))
            case FileType.LANG_KV:
                result = []
                # process all lang files and store them
                for lang_path in cwd.glob(self.path):
                    output_path = lang_path.parent / lang_path.name.replace(".lang", ".json")
                    _write_json_from_lang(lang_path, output_path)
                    result += output_path
                return result
            case FileType.FTBQuests_Chapter:
                result = []
                # make sure the output directory exists
                output_dir = cwd / "ftbquests"
                output_dir.mkdir(exist_ok=True)
                # process all chapter snbt files and store them
                for snbt_path in cwd.glob(self.path):
                    output_path = output_dir / snbt_path.name.replace(".snbt", ".json")
                    _write_json_from_ftbq_chapter_snbt(snbt_path, output_path)
                    result += output_path
                return result


@dataclass
class TranslatoolsMetadata:
    # the CurseForge project ID
    project_id: int
    # the Paratranz project ID
    paratranz_id: int
    # the CurseForge file ID, or 0 for unknown or uninitialized
    current_version_id: int = 0
    # the tracked flat key-value-paired JSON files, relative to the configuration file, glob supported
    # deprecated
    # TODO: REMOVE
    tracked_json_paths: list[str] | None = field(default=None)
    # the tracked Mojang-flavored LANG files, relative to the configuration file, glob supported
    # deprecated
    # TODO: REMOVE
    tracked_lang_paths: list[str] | None = field(default=None)
    # true to enable support for FTBQuests
    # but if somehow the modpack doesn't contain FTB Quests, or the contents are already localization-friendly,
    # set it to false.
    # deprecated
    # TODO: REMOVE
    ftbquests: bool | None = None
    # the tracked files
    tracked_files: list[TrackedFile] = field(default_factory=list)

    @staticmethod
    def load_from_path(path: Path) -> "TranslatoolsMetadata":
        dacite_conf = dacite.Config(type_hooks={FileType: FileType})
        return dacite.from_dict(TranslatoolsMetadata, json.loads(path.read_text(encoding="utf-8")), dacite_conf)

    @staticmethod
    def write_to_path(path: Path, config: "TranslatoolsMetadata"):
        path.write_text(json.dumps(dataclasses.asdict(config), indent=4), encoding="utf-8")


def _write_json_from_ftbq_chapter_snbt(snbt_path: Path, json_path: Path):
    json_content = _generate_json_from_ftbquests_chapter(snbt_path)
    json_path.write_text(json_content, encoding="utf-8")


def _write_json_from_lang(lang_path: Path, json_path: Path):
    json_content = _generate_json_from_lang(lang_path)
    json_path.write_text(json_content, encoding="utf-8")


def update_deprecated_metadata(config: TranslatoolsMetadata) -> bool:
    """
    Migrate deprecated fields, `tracked_json_paths`, `tracked_lang_paths` and `ftbquests`, to `tracked_files`, and
    remove the values, if capable.
    :param config: the config instance.
    :return: True if anything is changed in the config, which means you need to update the file.
    """
    updated = False

    if len(config.tracked_json_paths) > 0:
        for json_path in config.tracked_json_paths:
            config.tracked_files.append(TrackedFile(json_path, FileType.JSON_KV))
        config.tracked_json_paths.clear()
        print("'tracked_json_paths' is deprecated, and the paths are now migrated to new fields.")
        updated = True

    if len(config.tracked_lang_paths) > 0:
        for lang_path in config.tracked_lang_paths:
            config.tracked_files.append(TrackedFile(lang_path, FileType.LANG_KV))
        config.tracked_lang_paths.clear()
        print("'tracked_lang_paths' is deprecated, and the paths are now migrated to new fields.")
        updated = True

    if config.ftbquests:
        config.tracked_files.append(
            TrackedFile("**/config/ftbquests/quests/chapters/*.snbt", FileType.FTBQuests_Chapter))
        config.ftbquests = False
        print("'ftbquests' is deprecated, and the config is now migrated to new fields.")
        updated = True

    return updated


def _generate_json_from_lang(lang_path: Path) -> str:
    result = dict()
    with open(lang_path, encoding="utf-8") as lang_file:
        entries = lang_file.read().splitlines()
        for entry in entries:
            entry = entry.strip()
            # ignore comments
            if entry.startswith("#") or len(entry) == 0:
                continue
            pair = entry.split("=", maxsplit=1)
            if len(pair) != 2:
                raise ValueError(f"Unable to split the language entry: {entry}")
            # store the key-value pair
            result[pair[0]] = pair[1]

        return json.dumps(result, indent=4)


def _generate_json_from_ftbquests_chapter(snbt_path: Path,
                                          config: FTBQuestKeyGeneratingConfig = FTBQuestKeyGeneratingConfig.get_default()) -> str:
    """
    Generate key-value-paired JSON file from FTB Quests Chapter files.
    :param snbt_path: the SNBT file path
    :param config: the configuration
    :return: the result JSON file content
    """
    with open(snbt_path, encoding="utf-8") as f:
        data = ftb_snbt_lib.load(f)
        chapter_id = str(data["id"])
        quests: List = data["quests"]
        if not isinstance(quests, List):
            raise ValueError("'quests' in the snbt is not a list")

        result = dict()
        for quest in quests:
            quest: Compound
            quest_id = str(quest["id"])

            if "title" in quest:
                title = quest["title"]
                result[config.get_title_key(chapter_id, quest_id)] = str(title)
            if "subtitle" in quest:
                subtitle = quest["subtitle"]
                result[config.get_subtitle_key(chapter_id, quest_id)] = str(subtitle)
            if "description" in quest:
                description: List = quest["description"]
                count = 0
                for index, desc in enumerate(description):
                    if config.generate_description_index_for_empty_lines:
                        result[config.get_description_key(chapter_id, quest_id, index)] = str(desc)
                    else:
                        result[config.get_description_key(chapter_id, quest_id, count)] = str(desc)
                        count += 1

        return json.dumps(result, indent=4)



import argparse
import json
import os.path
import sys
from dataclasses import asdict
from pathlib import Path
from typing import overload, Literal

import cursefetch
from dotenv import load_dotenv

from translatools.config import TranslatoolsMetadata, update_deprecated_metadata, TrackedFile, FileType
from translatools.paratranz import Paratranz
from translatools.translatools import Translatools


def main() -> None:
    print("Hello from translatools!")

    load_dotenv()

    parser = argparse.ArgumentParser()
    parser.add_argument("--config", help="The configuration of the instance.", default="config.json")

    subparser = parser.add_subparsers(dest="command")

    # init
    init = subparser.add_parser("init", help="Initialize a new translation project from CurseForge packs.")
    init.add_argument("project_id", help="The CurseForge project ID.")
    init.add_argument("--version", help="The file id or name of the version to download.", default="latest")
    init.add_argument("-t", "--release-type",
                      help="The release type to filter by. (default: none). (only applicable when version is 'latest')",
                      choices=["release", "beta", "alpha"], default=None)
    init.add_argument("--api-key",
                      help="The CurseForge API key to use (can also be set via CF_API_KEY environment variable).")
    init.add_argument("--allow-non-empty-directory", action="store_true")

    # sync2paratranz
    sync2paratranz = subparser.add_parser("sync2paratranz")
    sync2paratranz.add_argument("--api-key",
                                help="The Paratranz API key to use (can also be set via PARATRANZ_API_KEY environment variable.)")

    # tracked
    tracked = subparser.add_parser("tracked")
    tracked_subparser = tracked.add_subparsers(dest="tracked_command")
    # tracked - add
    tracked_add = tracked_subparser.add_parser("add")
    tracked_add.add_argument("glob", help="The path or the glob to the files")
    tracked_add.add_argument("type", help="The type of tracked files")

    args = parser.parse_args()

    match args.command:
        case "init":
            _command_init(args)
        case "sync2paratranz":
            _command_sync_to_paratranz(args)
        case "tracked":
            _command_tracked(args)
        case _:
            parser.print_help()


@overload
def _get_translatools_from_args(args, exit_on_invalid_path: Literal[True]) -> Translatools: ...


@overload
def _get_translatools_from_args(args, exit_on_invalid_path: Literal[False]) -> Translatools | None: ...


def _get_translatools_from_args(args, exit_on_invalid_path: bool) -> Translatools | None:
    conf_path = Path(args.config)
    if not conf_path.exists():
        if exit_on_invalid_path:
            sys.exit(f"Configuration file is missing, expected {conf_path}")
        else:
            return None

    return Translatools(TranslatoolsMetadata.load_from_path(conf_path), conf_path)


def _command_init(args):
    config_path = Path(args.config)
    cwd = config_path.parent

    # check if the directory is clear
    if not args.allow_non_empty_directory and len(list(cwd.iterdir())) > 0:
        sys.exit("Expected a empty directory.")

    # download the modpack
    if args.api_key is not None:
        os.environ["CF_API_KEY"] = args.api_key
    f = cursefetch.get_project_file(str(args.project_id), args.version, args.release_type)
    cursefetch.download_project_file(f, ".", uncompress=True)

    # write the config
    conf = TranslatoolsMetadata(
        project_id=int(args.project_id),
        paratranz_id=0,
        current_version_id=f.id,
    )
    with open(config_path, "w+") as output:
        json.dump(asdict(conf), output, ensure_ascii=False, indent=4)
    print(f"Initialized project at {cwd}.")
    print("Modify the paratranz_id to start working with Paratranz.")


def _command_sync_to_paratranz(args):
    translatools_ = _get_translatools_from_args(args, True)
    para = Paratranz(os.environ.get("PARATRANZ_API_KEY", args.api_key))
    translatools_.sync_to_paratranz(para)


def _command_tracked(args):
    translatools_ = _get_translatools_from_args(args, True)
    config_path = translatools_.config_path()
    cwd = translatools_.cwd()

    # migrate!
    if update_deprecated_metadata(translatools_.config):
        TranslatoolsMetadata.write_to_path(config_path, translatools_.config)

    match args.tracked_command:
        case "add":
            glob = args.glob
            type_ = args.type

            if glob is None:
                sys.exit("Argument glob is missing")
            if type_ is None:
                sys.exit("Argument type is missing")
            try:
                FileType(type_)
            except ValueError:
                sys.exit("Argument type is invalid")

            tracked_file = TrackedFile(glob, type_)
            translatools_.config.tracked_files.append(tracked_file)
            TranslatoolsMetadata.write_to_path(config_path, translatools_.config)
            print(f"Added tracked: '{tracked_file.path}' as {tracked_file.type}")
            tracked_paths = "\n".join(f"- {p.absolute().as_posix()}" for p in tracked_file.get_paths(cwd))
            print(tracked_paths)
        case _:
            if len(translatools_.config.tracked_files) <= 0:
                print("Tracked nothing?! Add something by 'translatools tracked add'")
            else:
                for tracked_file in translatools_.config.tracked_files:
                    print(f"Tracked: '{tracked_file.path}' as {tracked_file.type}")
                    # the paths that matches
                    tracked_paths = "\n".join(f"- {p.absolute().as_posix()}" for p in tracked_file.get_paths(cwd))
                    print(tracked_paths)
                    # separator
                    print()
